#include <stdlib.h>
#include "mesh.h"
#include "primitives.h"

MeshRef MeshCreate(Vertex *vertices, GLshort *indices) {
  MeshRef self = (MeshRef) malloc(sizeof(Mesh));
  
  self->vertices = vertices;
  self->indices = indices;
  
  return self;
}

MeshRef MeshCreateGrid(int gridWidth, int gridHeight, GLfloat width, GLfloat height, GLfloat y) {
  int i, j, index;
  GLfloat horizontalStep = width / (gridWidth - 1);
  GLfloat verticalStep = height / (gridHeight - 1);
  GLfloat x = - width / 2.0, z = - height / 2.0;

  MeshRef self = (MeshRef) malloc(sizeof(Mesh));

  self->gridWidth = gridWidth; self->gridHeight = gridHeight;
  self->vertexCount = gridWidth * gridHeight;
  self->indexCount = (gridWidth*2 + 2) * (gridHeight - 1) - 2;
  
  self->vertices = (VertexRef) malloc(sizeof(Vertex) * self->vertexCount);
    
  index = 0;
  Vector normal = VectorMake(0.0, 1.0, 0.0);
  for(i = 0; i < gridHeight; i++) {
    for(j = 0; j < gridWidth; j++) {
      self->vertices[i*gridWidth + j] = VertexMake(PointMake(x + j * horizontalStep, y,  z + i * verticalStep), normal);
    }
  }
  
  self->indices = (GLshort *) malloc(sizeof(GLshort) * self->indexCount);
  
  index = 0;
  for(i = 0; i < gridHeight - 1; i++) {
    for(j = 0; j < gridWidth; j++) {
      self->indices[index++] = i * gridWidth + j;
      self->indices[index++] = (i + 1) * gridWidth + j;
    }
    
    if(i == gridHeight - 2) // Don't generate the degenerate triangles for the last row
      break;
    
    self->indices[index++] = (i + 2) * gridWidth - 1;
    self->indices[index++] = (i + 1) * gridWidth;
  }

  return self;
}

void MeshRecomputeNormals(MeshRef self) {
  int i, j;
  VertexRef v1, v2, aVertex;
  Vector u, v;
  
  for(i = 0; i < self->gridHeight; i++) {
    for(j = 0; j < self->gridWidth; j++) {
      aVertex = (self->vertices + i*self->gridWidth + j);
      
      v1 = (self->vertices + i * self->gridWidth + (j < self->gridWidth - 1 ? j + 1 : j));
      v2 = (self->vertices + i * self->gridWidth + (j > 0 ? j - 1 : j));
      u = VectorMake(v1->position[0] - v2->position[0], v1->position[1] - v2->position[1], v1->position[2] - v2->position[2]);
                         
      v1 = (self->vertices +  (i < self->gridHeight - 1 ? i + 1 : i) * self->gridWidth + j);
      v2 = (self->vertices + (i > 0 ? i - 1 : i) * self->gridWidth + j);
      v = VectorMake(v1->position[0] - v2->position[0], v1->position[1] - v2->position[1], v1->position[2] - v2->position[2]);
      
      // cross product
      Vector normal = CrossProduct(u, v);
      double length = VectorLength(normal);
      
      // printf("Computed normal: %f, %f, %f\n", normal[0], normal[1], normal[2]);
      
      aVertex->normal[0] = normal.u / length; 
      aVertex->normal[1] = normal.v / length;
      aVertex->normal[2] = normal.w / length;
    }
  }
  
}


void MeshDraw(MeshRef self) {
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  
  glVertexPointer(3, GL_FLOAT, sizeof(Vertex), self->vertices->position);
  glNormalPointer(GL_FLOAT, sizeof(Vertex), self->vertices->normal);
  glDrawElements(GL_TRIANGLE_STRIP, self->indexCount, GL_UNSIGNED_SHORT, self->indices);  
  
  glDisableClientState(GL_NORMAL_ARRAY);  
  glDisableClientState(GL_VERTEX_ARRAY);
}

