#include <stdlib.h>
#include <math.h>
#include "physics.h"

PointMass PointMassMake(VertexRef vertex) {
  PointMass result;
  
  result.vertex = vertex;
  result.lastPosition[0] = vertex->position[0];
  result.lastPosition[1] = vertex->position[1];
  result.lastPosition[2] = vertex->position[2];
  
  result.force = VectorMakeZero();
  
  return result;
}

// Verlet integration scheme
void PointMassIntegrate(PointMassRef self, double damp, double dt) {
  Vector velocity = VectorMake(self->vertex->position[0] - self->lastPosition[0], 
                               self->vertex->position[1] - self->lastPosition[1], 
                               self->vertex->position[2] - self->lastPosition[2]);
                                                                    
  self->lastPosition[0] = self->vertex->position[0];
  self->lastPosition[1] = self->vertex->position[1];
  self->lastPosition[2] = self->vertex->position[2];
  
  // Verlet step
  self->vertex->position[0] += velocity.u * (1.0 - damp) + (self->force.u)*dt*dt;
  self->vertex->position[1] += velocity.v * (1.0 - damp) + (self->force.v)*dt*dt;
  self->vertex->position[2] += velocity.w * (1.0 - damp) + (self->force.w)*dt*dt;
  
  self->force = VectorMakeZero();
}

// Spring constraint

Spring SpringMake(PointMassRef a, PointMassRef b, double stiffness) {
  Spring result;
  
  result.a = a;
  result.b = b;
  
  GLfloat *pa = a->vertex->position;
  GLfloat *pb = b->vertex->position;
  
  result.length = VectorLength(VectorMake(pb[0] - pa[0], pb[1] - pa[1], pb[2] - pa[2]));
  
  result.stiffness = stiffness;

  return result;
}

// doesn't really simulate forces, just moves points
void SpringSatisfy(SpringRef self) {
  GLfloat *pa = self->a->vertex->position;
  GLfloat *pb = self->b->vertex->position;
  
  Vector dv = VectorMake(pb[0] - pa[0], pb[1] - pa[1], pb[2] - pa[2]);
  double l = VectorLength(dv);
    
  dv.u /= l; dv.v /= l; dv.w /= l; // normalize dv
  l = self->length - l; // length difference

  if(fabs(l / self->length) < 0.005) // allow 0.5% length difference
    return;

  l *= self->stiffness;
  Vector force = VectorMake(dv.u * l, dv.v * l, dv.w * l);
  
  // move points (more stable than actually simulating forces)
  pa[0] -= force.u * 0.5; pa[1] -= force.v * 0.5; pa[2] -= force.w * 0.5;
  pb[0] += force.u * 0.5; pb[1] += force.v * 0.5; pb[2] += force.w * 0.5;
  
}

// Sphere constraint

Sphere SphereMake(Point center, double radius, double friction) {
  Sphere self;
  
  self.center = center;
  self.radius = radius;
  self.friction = friction;
  self.next = NULL;
  
  return self;
}


/* 
Now, this might be considered cheating. And it actually is cheating... The intersection of the particle
velocity vector with the sphere is computed almost as if the sphere was actually it's tangent plane.
This method provides a good enough result though, without causing me a headache and frying the CPU.
*/
void SphereSatisfyForPointMass(SphereRef self, PointMassRef pointMass) {
  Point c = self->center;
  GLfloat *v0 = pointMass->lastPosition;
  GLfloat *v1 = pointMass->vertex->position;

  Vector n1 = VectorMake(v1[0] - c.x, v1[1] - c.y, v1[2] - c.z);
  double l1 = VectorLength(n1);
  
  if(l1 > self->radius)// point is outside the sphere
    return;
  
  Vector n0 = VectorMake(v0[0] - c.x, v0[1] - c.y, v0[2] - c.z);
  double l0 = VectorLength(n1);

  l0 = self->radius / l0;  
  l1 = self->radius / l1;

  Point v0s = PointMake(c.x + n0.u*l0, c.y + n0.v*l0, c.z + n0.w*l0);
  Point v1s = PointMake(c.x + n1.u*l1, c.y + n1.v*l1, c.z + n1.w*l1);
  
  v1[0] = self->friction*v0s.x + (1.0 - self->friction)*v1s.x;
  v1[1] = self->friction*v0s.y + (1.0 - self->friction)*v1s.y;
  v1[2] = self->friction*v0s.z + (1.0 - self->friction)*v1s.z;
  
  return;
}

// Plane constraint

Plane PlaneMake(Vector normal, double shift) {
  Plane self;
  
  self.normal = normal;
  self.shift = shift;

  self.next = NULL;

  return self;
}

// one way plane, make sure the cloth starts outside the half-space!
void PlaneSatisfyForPointMass(PlaneRef self, PointMassRef pointMass) {
  GLfloat *p0 = pointMass->lastPosition;
  GLfloat *p1 = pointMass->vertex->position;
  
  Vector n = self->normal;
  double o = self->shift;
  
  if(n.u*p1[0] + n.v*p1[1] + n.w*p1[2] - o > 0.001)
    return;
  
  p1[0] = p0[0]; p1[1] = p0[1]; p1[2] = p0[2];
}

// Nail contraint

Nail NailMake(PointMassRef node) {
  Nail result;
  
  result.node = node;
  result.origin.x = node->vertex->position[0];
  result.origin.y = node->vertex->position[1];
  result.origin.z = node->vertex->position[2];
  
  result.next = NULL; // next in the list
  
  return result;
}

void NailSatisfy(NailRef self) {
  self->node->vertex->position[0] = self->origin.x;
  self->node->vertex->position[1] = self->origin.y;
  self->node->vertex->position[2] = self->origin.z;
}