#include "cloth.h"
#include <stdlib.h>

// private cloth methods

void ClothAccumulateForces(ClothRef self);
void ClothSatisfyConstraints(ClothRef self);

ClothRef ClothMake(MeshRef mesh, Vector gravity, double damp, double stretch, double shear, double bend) {
  ClothRef self = (ClothRef) malloc(sizeof(Cloth));
  
  self->spheres = NULL;
  self->planes = NULL;
  self->nails = NULL;
  self->mesh = mesh;
  self->width = mesh->gridWidth; // FIMXE redundant
  self->height = mesh->gridWidth; // FIMXE redundant
  self->dampingFactor = damp;
  self->gravity = gravity;
  
  long i, j, index;
  
  // initialize nodes
  self->nodeBuffer = (PointMassRef) malloc(sizeof(PointMass) * mesh->vertexCount);

  index = 0;
  for(j = 0; j < self->height; j++) {
    for(i = 0; i < self->width; i++) {
      self->nodeBuffer[index++] = PointMassMake(mesh->vertices + j*self->width + i);
    }
  }
    
  self->springCount = (self->width) * (self->height) * 2 - self->width - self->height; // stretch springs
  self->springCount += (self->width - 1) * (self->height - 1) * 2; // shear springs
  self->springCount += (self->width) * (self->height ) * 2 - 3*(self->width + self->height); // bend springs

  self->springBuffer = (SpringRef) malloc(sizeof(Spring) * self->springCount);
  
  // add springs
  index = 0;
  for(j = 0; j < self->height; j++) {
    for(i = 0; i < self->width; i++) {
      PointMassRef a = self->nodeBuffer + j*self->width + i;
      PointMassRef b, c;
            
      // horizontal stretch spring
      if(i < self->width - 1) {
        b = self->nodeBuffer + (j * self->width + (i + 1));
        self->springBuffer[index++] = SpringMake(a, b, stretch);
      }
      
      // vertical strech spring
      if(j < self->height - 1) {
        b = self->nodeBuffer + ((j + 1) * self->width + i);
        self->springBuffer[index++] = SpringMake(a, b, stretch);
      }
      
      // add shear springs
      if(i < self->width - 1 && j < self->height - 1) {
        b = self->nodeBuffer + ((j + 1) * self->width + (i+1));
        self->springBuffer[index++] = SpringMake(a, b, shear);

        b = self->nodeBuffer + (j * self->width + (i+1));
        c = self->nodeBuffer + ((j + 1) * self->width + i);
        self->springBuffer[index++] = SpringMake(b, c, shear);
      }

      // horizontal bend spring
      if(i < self->width - 3) {
        b = self->nodeBuffer + (j * self->width + (i + 3));
        self->springBuffer[index++] = SpringMake(a, b, bend);
      }
      
      // vertical bend spring
      if(j < self->height - 3) {
        b = self->nodeBuffer + ((j + 3) * self->width + i);
        self->springBuffer[index++] = SpringMake(a, b, bend);
      }
    }
  }
  
  return self;
} 

void ClothAddSphere(ClothRef self, SphereRef newSphere) {
  SphereRef sphere = self->spheres;
  
  if(sphere == NULL) {
    self->spheres = newSphere;
  } else {
    while(sphere->next)
      sphere = sphere->next;
    
    sphere->next = newSphere;
  }
}

void ClothAddPlane(ClothRef self, PlaneRef newPlane) {
  PlaneRef plane = self->planes;
  
  if(plane == NULL) {
    self->planes = newPlane;
  } else {
    while(plane->next)
      plane = plane->next;
    
    plane->next = newPlane;
  }
}

void ClothAddNail(ClothRef self, long x, long y) {
  PointMassRef node = self->nodeBuffer + (y*self->width + x);
  NailRef newNail = (NailRef) malloc(sizeof(Nail));
  *newNail = NailMake(node);
  
  NailRef nail = self->nails;
  if(nail == NULL) {
    self->nails = newNail;
  } else {
    while(nail->next)
      nail = nail->next;
    
    nail->next = newNail;
  }
}

void ClothStep(ClothRef self, double dt) {
  int i;
  
  ClothAccumulateForces(self);
    
  double damp = ((double) rand() / (double) RAND_MAX) * 0.01 + self->dampingFactor;
  
  for(i = 0; i < self->width * self->height; i++) {
    PointMassIntegrate(self->nodeBuffer + i, damp, dt);
  }
  
  ClothSatisfyConstraints(self);
  
  MeshRecomputeNormals(self->mesh);
}

// private functions

void ClothAccumulateForces(ClothRef self) {
  PointMassRef node;
  int index;
  for(index = 0; index < self->width * self->height; index++) {
    node = self->nodeBuffer + index;
    
    node->force.u += self->gravity.u;
    node->force.v += self->gravity.v;
    node->force.w += self->gravity.w;
  }
}

void ClothSatisfyConstraints(ClothRef self) {
  int index;
  
  // satisfy nails 
  NailRef nail = self->nails;
  while(nail) {
    NailSatisfy(nail);
    nail = nail->next;
  }
  
  // satisfy spheres
  SphereRef sphere = self->spheres;
  while(sphere) {
    for(index = 0; index < self->width * self->height; index++) {
      SphereSatisfyForPointMass(sphere, self->nodeBuffer + index);
    }
    
    sphere = sphere->next;
  }

  // satisfy planes
  PlaneRef plane = self->planes;
  while(plane) {
    for(index = 0; index < self->width * self->height; index++) {
      PlaneSatisfyForPointMass(plane, self->nodeBuffer + index);
    }
    
    plane = plane->next;
  }
  
  // relax all the springs
  for(index = 0; index < self->springCount; index++) {
    SpringSatisfy(self->springBuffer + index);
  }
}
